package org.codehaus.xfire.java;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;

import org.codehaus.xfire.MessageContext;
import org.codehaus.xfire.SOAPConstants;
import org.codehaus.xfire.fault.XFireFault;
import org.codehaus.xfire.java.mapping.TypeMapping;
import org.codehaus.xfire.java.mapping.TypeMappingRegistry;
import org.codehaus.xfire.service.SimpleService;
import org.codehaus.xfire.session.Session;

/**
 * @author <a href="mailto:dan@envoisolutions.com">Dan Diephouse</a>
 */
public class DefaultJavaService
    extends SimpleService
    implements JavaService
{
    private TypeMappingRegistry typeMappingRegistry;
    
    private TypeMapping typeMapping;
    
    private List allowedMethods;
    
    private Class serviceClass;
    
    private Hashtable operations;
    
    protected static final String SERVICE_CLASS = "serviceClass";

    protected static final String ALLOWED_METHODS= "allowedMethods";

    private boolean autoTyped = false;
    
    private int scope = JavaService.SCOPE_APPLICATION;

    /** The object if the scope is APPLICATION. */
    private Object appObj;
    
    public DefaultJavaService()
    {
        super();
        allowedMethods = new ArrayList();
        operations = new Hashtable();
    }

    /**
     * @param className
     * @param string
     */
    public void setServiceClass( String className )
        throws ClassNotFoundException
    {
        serviceClass = getClass().getClassLoader().loadClass( className );
        
        initializeOperations();
    }

    /**
     * @param serviceClass2
     */
    private void initializeOperations()
    {
        Method[] methods = serviceClass.getDeclaredMethods();
        
        // TODO: go through superclasses, stopping at Object.class
        
        for ( int i = 0; i < methods.length; i++ )
        {
            Method method = methods[i];
            
            String methodName = method.getName();
            
            int modifiers = method.getModifiers();
            if ( isAllowed( methodName )
                 &&
                 Modifier.isPublic(modifiers)
                 &&
                 !Modifier.isStatic(modifiers) )
            {
                addOperation( method );
            }
        }
    }

    /**
     * @param i
     * @param method
     * @param methodName
     */
    private void addOperation( Method method )
    {
        Operation op = new Operation(method, this);
        
        operations.put( method.getName(), op );
    }

    /**
     * Determines whether or not to expose the specified method.
     * 
     * @param methodName
     * @return
     */
    private boolean isAllowed(String methodName)
    {
        return ( allowedMethods.size() == 0 || allowedMethods.contains(methodName) );
    }

    /**
     * @see org.codehaus.xfire.java.JavaService#getOperation(java.lang.String, java.lang.String)
     */
    public Operation getOperation( String localName, String namespace )
    {
        return (Operation) operations.get( localName );
    }

    /**
     * @see org.codehaus.xfire.java.JavaService#getOperations()
     */
    public Collection getOperations()
    {
        return operations.values();
    }
    
    /**
     * @return Returns the allowedMethods.
     */
    public List getAllowedMethods()
    {
        return allowedMethods;
    }

    /**
     * @param allowedMethods The allowedMethods to set.
     */
    public void setAllowedMethods(List allowedMethods)
    {
        this.allowedMethods = allowedMethods;
    }

    /**
     * @return Returns the typeMapping.
     */
    public TypeMapping getTypeMapping()
    {
        return typeMapping;
    }

    /**
     * @param typeMapping The typeMapping to set.
     */
    public void setTypeMapping(TypeMapping typeMapping)
    {
        this.typeMapping = typeMapping;
    }

    /**
     * @return
     */
    public Class getServiceClass()
    {
        return serviceClass;
    }

    /**
     * Load a class from the class loader.
     * 
     * @param className The name of the class.
     * @return The class.
     * @throws Exception
     */
    protected Class loadClass( String className )
        throws Exception
    {
        // Handle array'd types.
        if ( className.endsWith("[]") )
        {
            className = "[L" + className.substring(0, className.length() - 2 ) + ";";
        }
        
        return getClass().getClassLoader().loadClass( className );
    }
    
    /**
     * @see org.apache.avalon.framework.activity.Initializable#initialize()
     */
    public void initialize() throws Exception
    {
        TypeMapping tm = null;
        
        // If we are a enocded, use the encoded type mapping, else just use XSD
        if ( getUse() != null )
        {            
	        if ( getUse().equals(SOAPConstants.USE_ECNODED) )
	        {
	            tm = getTypeMappingRegistry().createTypeMapping( SOAPConstants.SOAP11_ENCODED, autoTyped );
	        }
	        else
	        {
	            tm = getTypeMappingRegistry().createTypeMapping( SOAPConstants.XSD, autoTyped );
	        }
	        setTypeMapping( tm );
	        getTypeMappingRegistry().register( getDefaultNamespace(), tm );
       }
    }

    /**
     * Creates and returns a service object depending on the
     * scope.
     */
    public Object getServiceObject(MessageContext context) 
    	throws XFireFault
	{
        if ( scope == JavaService.SCOPE_APPLICATION )
        {
            if ( appObj == null )
            {
                synchronized( DefaultJavaService.class )
                {
                    appObj = createServiceObject();
                }
            }
            return appObj;
        }
        else if ( scope == JavaService.SCOPE_SESSION )
        {
            Session session = context.getSession();
            String key = "service."+getName();
            
            Object sessObj = session.get(key);
            if ( sessObj == null )
            {
                synchronized( DefaultJavaService.class )
                {
	                sessObj = createServiceObject();
	                session.put(key, sessObj);
                }
            }
            return sessObj;
        }
        else if ( scope == JavaService.SCOPE_REQUEST )
        {
            return createServiceObject();
        }
        else
        {
            throw new UnsupportedOperationException("Scope " + scope + " is invalid.");
        }
	}
    
    public Object createServiceObject() throws XFireFault
    {
        try
	    {
	        return getServiceClass().newInstance();
	    }
	    catch (InstantiationException e)
	    {
	        throw new XFireFault( "Couldn't instantiate service object.", e, XFireFault.RECEIVER );
	    }
	    catch (IllegalAccessException e)
	    {
	        throw new XFireFault( "Couldn't access service object.", e, XFireFault.RECEIVER );
	    }
    }

	public boolean isAutoTyped()
	{
		return autoTyped;
	}
    
	public void setAutoTyped(boolean autoTyped)
	{
		this.autoTyped = autoTyped;
	}
 
    public TypeMappingRegistry getTypeMappingRegistry()
    {
        return typeMappingRegistry;
    }
    
    public void setTypeMappingRegistry(TypeMappingRegistry typeMappingRegistry)
    {
        this.typeMappingRegistry = typeMappingRegistry;
    }
    
    public int getScope()
    {
        return scope;
    }

    public void setScope(int scope)
    {
        this.scope = scope;
    }
}
